/*
 * LED interface for Himalaya, the HTC PocketPC.
 *
 * License: GPL
 *
 * Author: Luke Kenneth Casson Leighton, Copyright(C) 2004
 *
 * Copyright(C) 2004, Luke Kenneth Casson Leighton.
 *
 * History:
 *
 * 2004-02-19	Luke Kenneth Casson Leighton	created.
 *
 */
 
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/soc/asic3_base.h>

#include <asm/hardware/ipaq-asic3.h>
#include <asm/hardware/ipaq-ops.h>
#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach-types.h>
#include <asm/arch/ipaq.h>
#include <asm/arch/htcblueangel-gpio.h>
#include <asm/arch/htcblueangel-asic.h>

#include "../himalaya/tsc2200.h"

#ifdef DEBUG
#define dprintk(x...) printk(x)
#else
#define dprintk(x...)
#endif

extern struct platform_device blueangel_asic3;

enum button_irq_type {
	gpio, asic3d, tsc2200
};

#define FN_BIT 0x80
#define SHIFT_BIT 0x80

static struct qkey {
	unsigned char code;
	unsigned char key;
	
} qkeys[] = {
	{0x01, KEY_ESC},	/* operator button */
	{0x05, KEY_L},
	{0x06, KEY_O},
	{0x07, KEY_P},
	{0x08, KEY_ENTER},
	{0x12, KEY_LEFTSHIFT},
	{0x14, KEY_RESERVED},	/* FN */
	{0x17, KEY_MINUS},	/* ? */
	{0x23, KEY_TAB},
	{0x26, KEY_DOT},	/* . */
	{0x27, KEY_SLASH},	/* _ */
	{0x28, KEY_LEFTCTRL},	/* Menu */
	{0x34, KEY_BACKSPACE},
	{0x35, KEY_T},
	{0x36, KEY_Y},
	{0x37, KEY_U},
	{0x38, KEY_I},
	{0x41, KEY_Q},
	{0x42, KEY_W},
	{0x43, KEY_E},
	{0x44, KEY_R},
	{0x45, KEY_G},
	{0x46, KEY_H},
	{0x47, KEY_J},
	{0x48, KEY_K},
	{0x51, KEY_A},
	{0x52, KEY_S},
	{0x53, KEY_D},
	{0x54, KEY_F},
	{0x55, KEY_V},
	{0x56, KEY_B},
	{0x57, KEY_N},
	{0x58, KEY_M},
	{0x61, KEY_Z},
	{0x62, KEY_X},
	{0x63, KEY_C},
	{0x64, KEY_LEFTALT},
	{0x65, KEY_SPACE},
	{0x67, KEY_COMMA},	/* , */
	/* With FN pressed */
	{0x01|FN_BIT, KEY_ESC},	/* vodafone button */
	{0x05|FN_BIT, KEY_6},
	{0x06|FN_BIT, KEY_3},
	{0x07|FN_BIT, KEY_MINUS|SHIFT_BIT}, /* sharp s */
	{0x08|FN_BIT, KEY_ENTER},
	{0x12|FN_BIT, KEY_CAPSLOCK},
	{0x14|FN_BIT, KEY_RESERVED},	/* FN */
	{0x17|FN_BIT, KEY_9},		/* ? */
	{0x23|FN_BIT, KEY_TAB},
	{0x26|FN_BIT, KEY_0},
	{0x27|FN_BIT, KEY_BACKSLASH},	/* _ */
	{0x28|FN_BIT, KEY_LEFTCTRL},	/* Menu */
	{0x34|FN_BIT, KEY_BACKSPACE},
	{0x35|FN_BIT, KEY_5|SHIFT_BIT},
	{0x36|FN_BIT, KEY_6|SHIFT_BIT},
	{0x37|FN_BIT, KEY_1},
	{0x38|FN_BIT, KEY_2},
	{0x41|FN_BIT, KEY_1|SHIFT_BIT},
	{0x42|FN_BIT, KEY_2|SHIFT_BIT},
	{0x43|FN_BIT, KEY_3|SHIFT_BIT},
	{0x44|FN_BIT, KEY_4|SHIFT_BIT},
	{0x45|FN_BIT, KEY_UP},
	{0x46|FN_BIT, KEY_RIGHTBRACE},	/* + */
	{0x47|FN_BIT, KEY_4},
	{0x48|FN_BIT, KEY_5},
	{0x51|FN_BIT, KEY_8|SHIFT_BIT},
	{0x52|FN_BIT, KEY_9|SHIFT_BIT},
	{0x53|FN_BIT, KEY_7|SHIFT_BIT},
	{0x54|FN_BIT, KEY_0|SHIFT_BIT},
	{0x55|FN_BIT, KEY_DOWN},
	{0x56|FN_BIT, KEY_RIGHT},
	{0x57|FN_BIT, KEY_7},
	{0x58|FN_BIT, KEY_8},
	{0x61|FN_BIT, KEY_RESERVED},	/* @ */
	{0x62|FN_BIT, KEY_RESERVED},	/* Euro Sign */
	{0x63|FN_BIT, KEY_LEFT},
	{0x64|FN_BIT, KEY_LEFTALT},
	{0x65|FN_BIT, KEY_SPACE},
	{0x67|FN_BIT, KEY_RIGHTBRACE|SHIFT_BIT},	/* * */
};

static int num_qkeys=sizeof(qkeys)/sizeof(struct qkey);

static struct key {
	char *name;
	enum button_irq_type irq_type;
	int irq;
	int irq_level;
	int key;
} keys[] = {
	{"power_button",	gpio,	  IRQ_NR_BLUEANGEL_POWER_BUTTON, IRQT_FALLING, KEY_POWER},
	{"menu_button",		asic3d,   5,				 IRQT_FALLING, KEY_F3 /* KEY_MENU */},
	{"ok_button",		asic3d,   7,				 IRQT_FALLING, KEY_F4 /* KEY_OK */},
	{"windows_button",	asic3d,   8,				 IRQT_FALLING, KEY_F1},
	{"record_button",	asic3d,   9,				 IRQT_FALLING, KEY_F9 /* KEY_RECORD */},
	{"camera_button",	asic3d,  10,				 IRQT_FALLING, KEY_F10 /* KEY_CAMERA */},
	{"volume_slider_up",	asic3d,  11,				 IRQT_FALLING, KEY_VOLUMEUP},
	{"volume_slider_down",	asic3d,  12,				 IRQT_FALLING, KEY_VOLUMEDOWN},
	{"mail_button",		asic3d,  15,				 IRQT_FALLING, KEY_F2 /* KEY_EMAIL */},
	{"contacts",		tsc2200,  0,				 IRQT_RISING,  KEY_F5 /* KEY_CONTACTS */},
	{"calendar",		tsc2200,  1,				 IRQT_RISING,  KEY_F6 /* KEY_CALENDAR */},
	{"up",			tsc2200,  3,				 IRQT_RISING,  KEY_UP},
	{"select",		tsc2200,  6,				 IRQT_RISING,  KEY_KPENTER /* KEY_SELECT */},
	{"left",		tsc2200,  7,				 IRQT_RISING,  KEY_LEFT},
	{"phone_lift",		tsc2200,  9,				 IRQT_RISING,  KEY_F7 /* KEY_PHONE */},
	{"down",		tsc2200,  11,				 IRQT_RISING,  KEY_DOWN },
	{"phone_hangup",	tsc2200,  12,				 IRQT_RISING,  KEY_F8 /* CANCEL */},
	{"right",		tsc2200,  15,				 IRQT_RISING,  KEY_RIGHT},
};

static int num_keys=sizeof(keys)/sizeof(struct key);

static struct timer_list timer;

#define KP_POLL_TIME 20 /* milliseconds */

extern struct platform_device blueangel_asic3;

static struct input_dev blueangel_kp_input_dev;

static void
blueangel_kp_report_key(struct key *k, int pressed)
{
	input_report_key(&blueangel_kp_input_dev, k->key, pressed);
	input_sync(&blueangel_kp_input_dev);

	if (pressed)
		dprintk("%s pressed\n", k->name);
	else
		dprintk("%s released\n", k->name);
}

static int asic3d_old;
static int tsc2200_old;
static int fn_bit;

static void
blueangel_kp_check_keys(int irq_type, int *old, int new)
{
	int i;
	int xor;
	struct key *k;

	xor=new ^ *old;
	for (i = 0 ; i < num_keys ; i++) {
		k=&keys[i];
		if (k->irq_type == irq_type && (xor & (1 << k->irq))) 
			blueangel_kp_report_key(k, new & (1 << k->irq));
	}
	*old=new;
}

static irqreturn_t blueangel_qkbd_irq(int irq, void *data, struct pt_regs *regs)
{
	int timeout=0,key,dummy,pressed,i,keysym,shift;
	dprintk("%s: got interrupt for %d...\n", __FUNCTION__, irq);
	input_regs(&blueangel_kp_input_dev, regs);

	dummy=asic3_read_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_TxData);
	asic3_write_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_TxData, 0);
	asic3_write_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Control, 
		asic3_read_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Control) | 0x8000);
	asic3_write_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Control, 
		asic3_read_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Control) | 0x10);
	for (;;) {
		if (!(asic3_read_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Status) & 0x10))
			break;
		udelay(10);
		if (timeout++ > 10) {
			printk("Timeout reading key\n");
			break;
		}
	}
	key=asic3_read_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_RxData);
	
	pressed=(key & 0x80) == 0;
	key&=0x7f;
	key|=fn_bit;
	dprintk("received key=0x%x\n", key);
	for (i = 0 ; i < num_qkeys ; i++) {
		if (qkeys[i].code == key) {
			dprintk("reporting keysym %d pressed=%d\n", qkeys[i].key, pressed);
			if ((key & ~FN_BIT) == 0x14) {
				fn_bit=pressed ? FN_BIT : 0;
			} else {
				keysym=qkeys[i].key;
				shift=keysym & SHIFT_BIT;
				keysym &= ~SHIFT_BIT;
				if (shift && pressed) {
					input_report_key(&blueangel_kp_input_dev, KEY_RIGHTSHIFT, 1);
					input_sync(&blueangel_kp_input_dev);
				}
				input_report_key(&blueangel_kp_input_dev, keysym, pressed);
				input_sync(&blueangel_kp_input_dev);
				if (shift && !pressed) {
					input_report_key(&blueangel_kp_input_dev, KEY_RIGHTSHIFT, 0);
					input_sync(&blueangel_kp_input_dev);
				}
			}
			return IRQ_HANDLED;
		}
	}
	printk("key unhandled\n");
        return IRQ_HANDLED;
}

static irqreturn_t blueangel_kp_irq(int irq, void *data, struct pt_regs *regs)
{
	struct key *k=data;
	dprintk("%s: got interrupt for %d...\n", __FUNCTION__, irq);
	input_regs(&blueangel_kp_input_dev, regs);
	if (k->irq_level == IRQT_FALLING)
		k->irq_level=IRQT_RISING;
	else
		k->irq_level=IRQT_FALLING;
	set_irq_type(irq, k->irq_level);
	if (k->irq_type == asic3d) 
		blueangel_kp_check_keys(asic3d, &asic3d_old, ~(ipaq_asic3_read_gpio_status_d(&blueangel_asic3.dev)));
	else
		blueangel_kp_report_key(k, !(GPLR(GPIO_NR_BLUEANGEL_POWER_BUTTON_N) & GPIO_bit(GPIO_NR_BLUEANGEL_POWER_BUTTON_N)));
        return IRQ_HANDLED;
}

static void
blueangel_kp_timer(unsigned long dummy)
{
	int status, keys;

	status=tsc2200_read(TSC2200_CTRLREG_KEY);
	if (status & 0x8000) 
		keys=tsc2200_read(TSC2200_DATAREG_KPDATA);
	else
		keys=0;
	blueangel_kp_check_keys(tsc2200,&tsc2200_old, keys);
	if (keys)
		mod_timer (&timer, jiffies + (KP_POLL_TIME * HZ) / 1000);
	else
		enable_irq(IRQ_NR_BLUEANGEL_TSC2200_KB);
}

static irqreturn_t blueangel_kp_tsc2200_irq(int irq, void *data, struct pt_regs *regs)
{
	int status, keys;
	input_regs(&blueangel_kp_input_dev, regs);
	status=tsc2200_read(TSC2200_CTRLREG_KEY);
	if (status & 0x8000) 
		keys=tsc2200_read(TSC2200_DATAREG_KPDATA);
	else
		keys=0;
	blueangel_kp_check_keys(tsc2200,&tsc2200_old, keys);
	if (keys) {
		dprintk("keys=0x%x\n", keys);
		disable_irq(IRQ_NR_BLUEANGEL_TSC2200_KB);
		mod_timer (&timer, jiffies + (KP_POLL_TIME * HZ) / 1000);
	}
        return IRQ_HANDLED;
}

static int blueangel_kp_init (void)
{
	int i,irq,key;
	struct key *k;

	dprintk("%s\n", __FUNCTION__);
	disable_irq(ASIC3_SPI_IRQ);
	asic3_write_register(&blueangel_asic3.dev, _IPAQ_ASIC3_SPI_Base+_IPAQ_ASIC3_SPI_Int, 0);

        blueangel_kp_input_dev.evbit[0] = BIT(EV_KEY) | BIT(EV_REP);
	init_input_dev(&blueangel_kp_input_dev);
	for (i = 0 ; i < num_keys ; i++) {
		k=&keys[i];
		irq=-1;
                set_bit(k->key, blueangel_kp_input_dev.keybit);
		switch (k->irq_type) {
		case gpio:
			irq=k->irq;
			break;
		case asic3d:
			irq=asic3_irq_base(&blueangel_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + k->irq;
			break;
		case tsc2200:
			break;
		}
		if (irq != -1) {
			dprintk("requesting irq %d\n", irq);
			request_irq(irq, blueangel_kp_irq, SA_SAMPLE_RANDOM, k->name, k);
			set_irq_type(irq, k->irq_level);
		}
#if 0
		blueangel_kp_input_dev.keybit[k->key] |= BIT(k->key);
#endif
	}
	for (i = 0 ; i < num_qkeys ; i++) {
		key=qkeys[i].key;

		if (key) {
			dprintk("enabling key %d\n", qkeys[i].key);
			set_bit(key, blueangel_kp_input_dev.keybit);
		}
	}

	request_irq(IRQ_NR_BLUEANGEL_TSC2200_KB, blueangel_kp_tsc2200_irq, SA_SAMPLE_RANDOM, "blueangel_tsc2200_kb", NULL);
	set_irq_type(IRQ_NR_BLUEANGEL_TSC2200_KB, IRQT_FALLING);
	irq=asic3_irq_base(&blueangel_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + GPIOD_QKBD_IRQ;
	request_irq(irq, blueangel_qkbd_irq, SA_SAMPLE_RANDOM, "blueangel_qkbd", NULL);
	set_irq_type(irq, IRQT_FALLING);


	init_timer(&timer);
	timer.function = blueangel_kp_timer;
	timer.data = 0;


        blueangel_kp_input_dev.name = "blueangelkb";
        blueangel_kp_input_dev.phys = "keyboard/blueangelkb";

//        blueangel_kp_input_dev.id.bustype = BUS_PXA_NSSP;
        blueangel_kp_input_dev.id.vendor = 0x0001;
        blueangel_kp_input_dev.id.product = 0x0001;
        blueangel_kp_input_dev.id.version = 0x0100;

        input_register_device(&blueangel_kp_input_dev);


	return 0;
}

static void blueangel_kp_exit (void)
{
	int i,irq;
	struct key *k;
	
	dprintk("%s\n", __FUNCTION__);
	input_unregister_device(&blueangel_kp_input_dev);
	for (i = 0 ; i < num_keys ; i++) {
		k=&keys[i];
		irq=-1;
		switch (k->irq_type) {
		case gpio:
			irq=k->irq;
			break;
		case asic3d:
			irq=asic3_irq_base(&blueangel_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + k->irq;
			break;
		case tsc2200:
			break;
		}
		if (irq != -1) {
			dprintk("releasing irq %d\n", irq);
			free_irq(irq, k);
		}
	}
	free_irq(IRQ_NR_BLUEANGEL_TSC2200_KB, NULL);
	irq=asic3_irq_base(&blueangel_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + GPIOD_QKBD_IRQ;
	free_irq(irq, NULL);
	del_timer_sync (&timer);
}

module_init (blueangel_kp_init);
module_exit (blueangel_kp_exit);

