/*
 * linux/arch/arm/mach-pxa/htcunivesal/htcuniversal_kp.c
 *
 * Keypad interface for Universal, the HTC PocketPC.
 *
 * License: GPL
 *
 * Author: Milan Votava, Copyright(C) 2006
 *
 * Based on:
 * linux/arch/arm/mach-pxa/blueangel/blueangel_kp.c
 * Copyright(C) 2004, Luke Kenneth Casson Leighton.
 *
 *
 * linux/drivers/misc/keypad.c
 * Copyright (C) 2004 Motorola
 *
 *
 * History:
 *
 * 2006-02-28	Milan Votava	created.
 *
 */
 
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/soc/asic3_base.h>

#include <asm/hardware/ipaq-asic3.h>
#include <asm/hardware/ipaq-ops.h>
#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/mach-types.h>

#include <asm/arch/htcuniversal-gpio.h>
#include <asm/arch/htcuniversal-asic.h>

/* keypad PXA27x registry macros */

#define KPAS_MUKP	0x7c000000 /* KPAS bits 30-26 */
#define MUKP_NO_KEYS	0x00000000 /* value of MUKP if no keys are down */
#define MUKP_ONE_KEY	0x04000000 /* value of MUKP if one key is down */
#define KPAS_RP		0x000000f0  /* row of single key    */
#define KPAS_CP		0x0000000f  /* column of single key */

#define NUM_WORDS_IN_BITMAP 2

#define KEYDOWN         0x8000
#define KEYUP           0x0000

#define RDELAY 200

//#define USE_RELEASE_TIMER

#ifdef DEBUG
#define dprintk(x...) printk(x)
#else
#define dprintk(x...)
#endif

extern struct platform_device htcuniversal_asic3;

#ifdef USE_RELEASE_TIMER
static void release_timer_went_off (unsigned long);
static void do_scan(void);
#endif

enum button_irq_type {
	asic3a, asic3b, asic3c, asic3d
};

typedef struct key {
	char *name;
	enum button_irq_type irq_type;
	int irq;
	int irq_level;
	int code;
	void (*handler)(int, struct key *);
} key;


static void autorepeat_timer_went_off (unsigned long);

static int add_events (unsigned long *, unsigned long *);
static void scan_to_bitmap (unsigned long, unsigned long *, unsigned long *);
static inline void set_bitmap_to_zero(unsigned long b[]);
static inline void copy_bitmap(unsigned long dest[], unsigned long source[]);
static inline void turn_on_bit(unsigned long map[], int);
static inline void turn_off_bit(unsigned long map[], int);
static inline int  any_keys_down (unsigned long *);
static int kp_input_report_key (int , int );
static void kp_handler_switch (int, struct key *);


static DECLARE_WAIT_QUEUE_HEAD(keypad_wait);
static spinlock_t keypad_lock = SPIN_LOCK_UNLOCKED;

/* previous bitmap of keys down */
static unsigned long oldkeybitmap[NUM_WORDS_IN_BITMAP];
/* current bitmap of keys down */
static unsigned long keybitmap[NUM_WORDS_IN_BITMAP];

static unsigned long kpas = 0;     /* last value of kpas */
static unsigned long kpasmkp[4] = {0, 0, 0, 0}; /* last value of kpasmkp0-3 */
static unsigned long kpdk = 0;     /* last value of kpdk */

#ifdef USE_RELEASE_TIMER
static struct timer_list key_release_timer;
static int key_release_interval = RDELAY * HZ/1000;
#endif

/* autorepeat done at input subsystem */
static int  do_autorepeat = 0;

static long jiffies_to_first_repeat = 30;
static long jiffies_to_next_repeat = 30;
static struct timer_list autorepeat_timer;

extern int lockscreen_flag;

#define FN_KEY 0x39
#define FN_BIT 0x80
#define SHIFT_BIT 0x80
#define FLAG_KEY_PRESSED 0x1

#define FN_KEY_OFFSET 0x40

static struct qkey {
	unsigned char code;
	unsigned char flags;
	unsigned short key;
	
} qkeys[] = {
	{0x01, 0, KEY_ENTER},	/* ENTER OK */
	{0x02, 0, KEY_COMMA},	/* , OK */
	{0x03, 0, KEY_UP},	/* UP OK */
	{0x04, 0, KEY_DOT},	/* . OK  */
	{0x05, 0, KEY_F9 /* Qt::Key_Yes KEY_OK*/},		/* OK */
	{0x06, 0, KEY_RIGHT},	/* RIGHT OK */ 
	{0x07, 0, KEY_F9 /* Qt::Key_Yes KEY_PHONE*/},	/* phone_lift (hinge) */
	{0x08, 0, KEY_F10 /* Qt::Key_No KEY_CANCEL*/},	/* phone_hangup (hinge) */
	{0x09, 0, KEY_MINUS},	/* - OK */
	{0x0a, 0, KEY_EQUAL},	/* = OK */
	{0x0b, 0, KEY_I},	/* OK */
	{0x0c, 0, KEY_O},	/* OK */
	{0x0d, 0, KEY_P},	/* OK */
	{0x0e, 0, KEY_BACKSPACE}, /* OK */
	{0x0f, 0, KEY_K},	/* OK */
	{0x10, 0, KEY_L},	/* OK */
	{0x11, 0, KEY_ESC /*KEY_F7|KEY_PHONE*/},	/* phone_lift */
	{0x12, 0, KEY_F1},	/* left soft button */
	{0x13, 0, KEY_F2 /*KEY_CONTACTS*/},	/* contacts */
	{0x14, 0, KEY_F3 /*KEY_VIDEO*/},	/* video call */
	{0x15, 0, KEY_F4},	/* right soft button */
	{0x16, 0, KEY_F5 /*KEY_CANCEL*/},	/* phone_hangup */
	{0x17, 0, KEY_9},	/* OK */
	{0x18, 0, KEY_0},	/* OK */
	{0x19, 0, KEY_1},	/* OK */
	{0x1a, 0, KEY_2},	/* OK */
	{0x1b, 0, KEY_3},	/* OK */
	{0x1c, 0, KEY_4},	/* OK */
	{0x1d, 0, KEY_5},	/* OK */
	{0x1e, 0, KEY_6},	/* OK */
	{0x1f, 0, KEY_7},	/* OK */
	{0x20, 0, KEY_8},	/* OK */
	{0x21, 0, KEY_TAB},	/* OK */
	{0x22, 0, KEY_Q},	/* OK */
	{0x23, 0, KEY_W},	/* OK */
	{0x24, 0, KEY_E},	/* OK */
	{0x25, 0, KEY_R},	/* OK */
	{0x26, 0, KEY_T},	/* OK */
	{0x27, 0, KEY_Y},	/* OK */
	{0x28, 0, KEY_U},	/* OK */
	{0x29, 0, KEY_CAPSLOCK},	/* OK */
	{0x2a, 0, KEY_A},	/* OK */
	{0x2b, 0, KEY_S},	/* OK */
	{0x2c, 0, KEY_D},	/* OK */
	{0x2d, 0, KEY_F},	/* OK */
	{0x2e, 0, KEY_G},	/* OK */
	{0x2f, 0, KEY_H},	/* OK */
	{0x30, 0, KEY_J},	/* OK */
	{0x31, 0, KEY_LEFTSHIFT},	/* OK */
	{0x32, 0, KEY_Z},	/* OK */
	{0x33, 0, KEY_X},	/* OK */
	{0x34, 0, KEY_C},	/* OK */
	{0x35, 0, KEY_V},	/* OK */
	{0x36, 0, KEY_B},	/* OK */
	{0x37, 0, KEY_N},	/* OK */
	{0x38, 0, KEY_M},	/* OK */
	{0x39, 0, KEY_RESERVED},	/* FN */
	{0x3a, 0, KEY_LEFTCTRL /*KEY_EMAIL*/},	/* messaging OK */
	{0x3b, 0, KEY_F6 /*KEY_HOME*/},	/* windows start button */
//	{0x3c, 0, KEY_F7 /*KEY_WWW*/},	/* explorer */
	{0x3c, 0, KEY_LEFTALT /*KEY_WWW*/},	/* explorer OK */
	{0x3d, 0, KEY_SPACE},	/* OK */
	{0x3e, 0, KEY_F7 /*Qt::Key_Menu KEY_MENU*/},	/* Sym */
	{0x3f, 0, KEY_LEFT},	/* LEFT OK */
	{0x40, 0, KEY_DOWN},	/* DOWN OK */


	/* With FN pressed */

	{0x02|FN_BIT, 0, KEY_SEMICOLON},	/* ; OK */
	{0x04|FN_BIT, 0, KEY_SEMICOLON|SHIFT_BIT},	/* : OK */
	{0x09|FN_BIT, 0, KEY_MINUS|SHIFT_BIT},	/* _ OK */
	{0x0a|FN_BIT, 0, KEY_EQUAL|SHIFT_BIT},  /* + OK */
	{0x0c|FN_BIT, 0, KEY_SLASH},		/* / OK */
	{0x0d|FN_BIT, 0, KEY_BACKSLASH},	/* \ OK */
	{0x0f|FN_BIT, 0, KEY_APOSTROPHE|SHIFT_BIT},	/* " OK */
	{0x10|FN_BIT, 0, KEY_APOSTROPHE},	/* ' OK */
	{0x11|FN_BIT, 0, KEY_F11 /*Qt::Key_Call |KEY_PHONE*/},	/* phone_lift */
	{0x13|FN_BIT, 0, KEY_F10 /*KEY_CONTACTS*/},	/* contacts */
	{0x16|FN_BIT, 0, KEY_F12 /*Qt::Key_Hangup KEY_CANCEL*/},	/* phone_hangup */
//	{0x17|FN_BIT, 0, KEY_RIGHTBRACE},
//	{0x18|FN_BIT, 0, KEY_LEFTBRACE},
	{0x17|FN_BIT, 0, KEY_9|SHIFT_BIT}, 	/* ( OK */
	{0x18|FN_BIT, 0, KEY_0|SHIFT_BIT}, 	/* ) OK */
	{0x19|FN_BIT, 0, KEY_1|SHIFT_BIT}, 	/* ! OK */
	{0x1a|FN_BIT, 0, KEY_2|SHIFT_BIT}, 	/* @ OK */
	{0x1b|FN_BIT, 0, KEY_3|SHIFT_BIT}, 	/* # OK */
	{0x1c|FN_BIT, 0, KEY_4|SHIFT_BIT}, 	/* $ OK */
	{0x1d|FN_BIT, 0, KEY_5|SHIFT_BIT}, 	/* % OK */
	{0x1e|FN_BIT, 0, KEY_6|SHIFT_BIT}, 	/* ^ OK */
	{0x1f|FN_BIT, 0, KEY_7|SHIFT_BIT}, 	/* & OK */
	{0x20|FN_BIT, 0, KEY_8|SHIFT_BIT}, 	/* * OK */
	{0x21|FN_BIT, 0, KEY_GRAVE|SHIFT_BIT},	/* ~ OK */
	{0x38|FN_BIT, 0, KEY_SLASH|SHIFT_BIT},	/* ? OK */
	{0x3a|FN_BIT, 0, KEY_F13 /*KEY_EMAIL*/},	/* messaging */

/* leave it as it is */
	{0x39|FN_BIT, 0, KEY_RESERVED},	/* FN */
};

static int num_qkeys=sizeof(qkeys)/sizeof(struct qkey);


static key keys[]={

/* ASIC3A */
	{"backlight_button",		asic3a,  GPIOA_BUTTON_BACKLIGHT_N,	IRQT_FALLING,  KEY_F10, NULL},
	{"record_button",		asic3a,  GPIOA_BUTTON_RECORD_N,	 	IRQT_FALLING,  KEY_F24 /*KEY_RECORD*/, NULL},
	{"camera_button",		asic3a,  GPIOA_BUTTON_CAMERA_N,	 	IRQT_FALLING,  KEY_F10 /*KEY_CAMERA*/, NULL},
	{"screen_cover",		asic3a,  GPIOA_COVER_ROTATE_N,	 	IRQT_FALLING, SW_0, kp_handler_switch},
	{"volume_slider_down",		asic3a,  GPIOA_VOL_UP_N,	 	IRQT_FALLING,  KEY_VOLUMEDOWN, NULL},
	{"volume_slider_up",		asic3a,  GPIOA_VOL_DOWN_N,	 	IRQT_FALLING,  KEY_VOLUMEUP, NULL},

/* ASIC3B */

	{"clamshell_rotate",		asic3b,  GPIOB_CLAMSHELL_N,	 IRQT_FALLING,  SW_1, kp_handler_switch},
	{"night_sensor",		asic3b,  GPIOB_NIGHT_SENSOR,	 IRQT_FALLING,  SW_2, kp_handler_switch},


/* ASIC3C */

/* ASIC3D */
	{"select",			asic3d,  GPIOD_KEY_OK_N,	 IRQT_FALLING,  KEY_KPENTER /*KEY_SELECT*/, NULL},
	{"right",			asic3d,  GPIOD_KEY_RIGHT_N,	 IRQT_FALLING,  KEY_RIGHT, NULL},
	{"left",			asic3d,  GPIOD_KEY_LEFT_N,	 IRQT_FALLING,  KEY_LEFT, NULL},
	{"down",			asic3d,  GPIOD_KEY_DOWN_N,	 IRQT_FALLING,  KEY_DOWN, NULL},
	{"up",				asic3d,  GPIOD_KEY_UP_N,	 IRQT_FALLING,  KEY_UP, NULL},
};

static int num_keys=sizeof(keys)/sizeof(struct key);

#define KP_POLL_TIME 20 /* milliseconds */

extern struct platform_device htcuniversal_asic3;

static struct input_dev *htcuniversal_kp_input_dev;

#ifdef USE_RELEASE_TIMER
/*
 * This is called when the key release timer goes off.  That's the
 * only time it should be called.  Check for changes in what keys
 * are down.
 */
static void
release_timer_went_off(unsigned long unused)
{
   unsigned long flags;

   spin_lock_irqsave(&keypad_lock,flags);
   do_scan();
   /* If any keys are still down, set the timer going again. */
   if (any_keys_down (keybitmap))
   {
      mod_timer (&key_release_timer, jiffies + key_release_interval);
   }
   else
   {
      del_timer (&autorepeat_timer);
   }
   spin_unlock_irqrestore(&keypad_lock, flags);
}
#endif

/*
 * This is called when the autorepeat timer goes off.
 */
static void
autorepeat_timer_went_off(unsigned long unused)
{
   int i, bitnum;
   unsigned long tmp;
   unsigned long flags;

   spin_lock_irqsave(&keypad_lock,flags);
   if (!any_keys_down (keybitmap))
   {
      spin_unlock_irqrestore(&keypad_lock, flags);
      return;
   }
   for (i=0;i<NUM_WORDS_IN_BITMAP;i++)
   {
     tmp = keybitmap[NUM_WORDS_IN_BITMAP-i-1];
     while ((bitnum = ffs(tmp)) != 0)
     {
       tmp &= ~(1<<(bitnum-1));
       kp_input_report_key (bitnum + (32*i), 1);
     }
   }
   spin_unlock_irqrestore(&keypad_lock, flags);
   wake_up_interruptible (&keypad_wait);

   mod_timer (&autorepeat_timer, jiffies + jiffies_to_next_repeat);
}

#ifdef USE_RELEASE_TIMER
/*
 * Do a scan and add key events for anything that's different from the
 * last scan.  We assume keypad_lock is locked when this is called
 * (by release_timer_went_off).
 */
static void
do_scan(void)
{
   unsigned long kpas_new, kpasmkp_new[4];
   unsigned long kpdk_new;
   int change;

   /* Initiate an automatic scan */
   KPC |= KPC_AS;
   /* Wait for the Scan On bit to go off before */
   /* reading the scan registers. */
   while (KPAS & KPAS_SO)
   { 
      NULL;
   };
   kpas_new = KPAS;
   kpasmkp_new[0] = KPASMKP0;
   kpasmkp_new[1] = KPASMKP1;
   kpasmkp_new[2] = KPASMKP2;
   kpasmkp_new[3] = KPASMKP3;
   kpdk_new = KPDK;
   /*
    * reading the register turns off the "key pressed since last read" bit
    * if it was on, so we turn it off 
    */
   kpdk_new &= ~KPDK_DKP;
   KPC &= ~KPC_AS;

   change = ((kpas_new != kpas) || 
             (kpasmkp_new[0] != kpasmkp[0]) || 
             (kpasmkp_new[1] != kpasmkp[1]) ||
             (kpasmkp_new[2] != kpasmkp[2]) ||
             (kpasmkp_new[3] != kpasmkp[3]) ||
             (kpdk_new != kpdk));

   if (change)
   {
      kpas = kpas_new;
      kpasmkp[0] = kpasmkp_new[0];
      kpasmkp[1] = kpasmkp_new[1];
      kpasmkp[2] = kpasmkp_new[2];
      kpasmkp[3] = kpasmkp_new[3];
      kpdk = kpdk_new;

      copy_bitmap (oldkeybitmap, keybitmap);
      scan_to_bitmap (kpas, kpasmkp, keybitmap);
      add_events (oldkeybitmap, keybitmap);

      wake_up_interruptible (&keypad_wait);
   }
}
#endif

/*
 * set all words in the specified bitmap to 0
 */
static inline void set_bitmap_to_zero(unsigned long b[]) {
   int i;
   for (i=0;i<NUM_WORDS_IN_BITMAP;i++)
   {
      b[i]=0;
   }
}

/*
 * copy source bitmap to destination bitmap
 */
static inline void copy_bitmap(unsigned long dest[], unsigned long source[]) {
   int i;
   for (i=0;i<NUM_WORDS_IN_BITMAP;i++)
   {
      dest[i]=source[i];
   }
}

/*
 * Turn on the bit position in map for the specified key code.
 * Bit numbers start with 1 (not 0) and increase moving left within
 * words and within the array.
 * Bit 1 is the rightmost bit in map[NUM_WORDS_IN_BITMAP-1].
 * Bit 33 is the rightmost bit in map[NUM_WORDS_IN_BITMAP-2].
 */
static inline void 
turn_on_bit(unsigned long map[], int code)
{
   int word, pos, p;

   /* if we have 2 words, bits 1-32 are in map[1], 33-64 in map[0] */
   word = (NUM_WORDS_IN_BITMAP - 1) - ((code-1) / 32);
   /* 
    * bit 1 is 1<<0, bit 2 is 1<<1, ... bit 32 is 1 << 31, all in map[1]
    * bit 33 is 1<<0, bit 34 is 1<<1, ... in map[0]
    * we want this:
    *     code       pos
    *      1          0
    *      2          1
    *          ...
    *     31         30
    *     32         31
    *     33          0
    *     34          1
    *          ...
    */
   pos = (p = (code % 32)) == 0 ? 31 : p-1;
   map[word] |= (1<<pos);
}

/*
 * Turn off the bit position in map for the specified key code.
 * Bit positions start with 1 (not 0).
 */
static inline void 
turn_off_bit(unsigned long map[], int code)
{
   int word, pos, p;

   word = (NUM_WORDS_IN_BITMAP - 1) - ((code-1) / 32);
   pos = (p = (code % 32)) == 0 ? 31 : p-1;
   map[word] &= ~(1<<pos);
}

/*
 * Return 1 if any bits are down in map[]
 */
static inline int
any_keys_down (unsigned long *map)
{
  int i;
  for (i=0;i<NUM_WORDS_IN_BITMAP;i++)
  {
    if (*map++)
    {
       return 1;
    }
  }
  return 0;
}

/*
 * Decode the specified scan register values into the bitmap pointed
 * to by the last argument.  The bitmap will contain a 1
 * for each key that's down.
 *
 * Only the 1st two of the four scan registers are used.
 */
static void
scan_to_bitmap(unsigned long kpas, unsigned long kpasmkp[], unsigned long bitmap[])
{
   int row, col;
   int bitnum;
   unsigned long scanmap;
   
   set_bitmap_to_zero (bitmap);

   if ((kpas & KPAS_MUKP) == MUKP_NO_KEYS)
   {
      return;
   }

   if ((kpas & KPAS_MUKP) == MUKP_ONE_KEY)
   {
      row = (kpas & KPAS_RP) >> 4;
      col = kpas & KPAS_CP;
      turn_on_bit (bitmap, col*8 + row+1);
      return;
   }

   dprintk("%04x.%04x.%04x.%04x\n", kpasmkp[0], kpasmkp[1], kpasmkp[2], kpasmkp[3]);
   /* reach here if multiple keys */
   scanmap = (kpasmkp[0] & 0xff) | ((kpasmkp[0] & 0xff0000) >> 8) |
             ((kpasmkp[1] & 0xff) << 16) | ((kpasmkp[1] & 0xff0000) << 8);
   while ((bitnum = ffs(scanmap)) != 0)
   {
     /*
      * ffs returns bit numbers starting with 1, so subtract 1 to index
      * into scanbit_to_keycode[]
      */
      turn_on_bit (bitmap, bitnum);
      scanmap &= ~(1<<(bitnum-1));
   }
   scanmap = (kpasmkp[2] & 0xff) | ((kpasmkp[2] & 0xff0000) >> 8) |
             ((kpasmkp[3] & 0xff) << 16) | ((kpasmkp[3] & 0xff0000) << 8);
   while ((bitnum = ffs(scanmap)) !=0)
   {
      turn_on_bit (bitmap, 32+bitnum);
      scanmap &= ~(1<<(bitnum-1));
   }   
}

/*
release all 'virtual' keys after the FN key is pressed up
*/

static void
kp_release_fn_keys()
{
	int i, keysym, shift;

	for (i = FN_KEY_OFFSET ; i < num_qkeys ; i++) {
		if (qkeys[i].flags | FLAG_KEY_PRESSED) {
			keysym=qkeys[i].key;
			shift=keysym & SHIFT_BIT;
			keysym &= ~SHIFT_BIT;
			if (shift) {
				input_report_key(htcuniversal_kp_input_dev, KEY_RIGHTSHIFT, 0);
				input_sync(htcuniversal_kp_input_dev);
			}
			input_report_key(htcuniversal_kp_input_dev, keysym, 0);
			input_sync(htcuniversal_kp_input_dev);
			qkeys[i].flags &= ~FLAG_KEY_PRESSED;
		}
	}
}

static int fn_bit;

static int 
kp_input_report_key (int scancode, int pressed)
{
	int i, keysym, shift, key;

	key=scancode | fn_bit;
	dprintk("received scancode=%x key=0x%x %x\n", scancode, key, pressed);
	for (i = 0 ; i < num_qkeys ; i++) {
		if (fn_bit && ! pressed && qkeys[i].code == scancode && qkeys[i].flags | FLAG_KEY_PRESSED) {
			keysym = qkeys[i].key;
			shift=keysym & SHIFT_BIT;
			keysym &= ~SHIFT_BIT;
			if (shift) {
				input_report_key(htcuniversal_kp_input_dev, KEY_RIGHTSHIFT, 0);
				input_sync(htcuniversal_kp_input_dev);
			}
			input_report_key(htcuniversal_kp_input_dev, keysym, 0);
			input_sync(htcuniversal_kp_input_dev);
			qkeys[i].flags &= ~FLAG_KEY_PRESSED;
		}
		if (qkeys[i].code == key) {
			dprintk("reporting keysym %d pressed=%d\n", qkeys[i].key, pressed);
			if (scancode == FN_KEY) {
				if (pressed)
					fn_bit=FN_BIT;
				else {
					fn_bit=0;
					kp_release_fn_keys();
				}
			} else {
				keysym=qkeys[i].key;
				shift=keysym & SHIFT_BIT;
				keysym &= ~SHIFT_BIT;
				if (shift && pressed) {
					dprintk("KEY_RIGHTSHIFT down\n");
					input_report_key(htcuniversal_kp_input_dev, KEY_RIGHTSHIFT, 1);
					input_sync(htcuniversal_kp_input_dev);
				}
				input_report_key(htcuniversal_kp_input_dev, keysym, pressed);
				input_sync(htcuniversal_kp_input_dev);
				if (shift && !pressed) {
					dprintk("KEY_RIGHTSHIFT up\n");
					input_report_key(htcuniversal_kp_input_dev, KEY_RIGHTSHIFT, 0);
					input_sync(htcuniversal_kp_input_dev);
				}
				input_sync(htcuniversal_kp_input_dev);
			}
			if (pressed)
				qkeys[i].flags |= FLAG_KEY_PRESSED;
			else
				qkeys[i].flags &= ~FLAG_KEY_PRESSED;
			
			return 1;
		}
	}
//	printk("key unhandled\n");
	return 0;
}


/* 
 * Add events indicated by the difference between the last scan (oldbitmap)
 * and this one (newbitmap) to the input buffer.
 * Return nonzero right away if any of the events can't be added.
 * A zero return means all the events were added.
 * This should be called with keypad_lock locked.
 */
static int
add_events (unsigned long *oldbitmap, unsigned long *newbitmap)
{
   unsigned long tmpmap, onebitmap;
   int bitnum, i;

   /* generate events for keys that were down before and are up now */
   for (i=NUM_WORDS_IN_BITMAP-1;i>=0;i--)
   {
     tmpmap = oldbitmap[i];
     while ((bitnum = ffs(tmpmap)) != 0)
     {
       onebitmap = 1<<(bitnum-1);
       if ((newbitmap[i] & onebitmap) == 0) 
       {
	 kp_input_report_key(bitnum + 32*(NUM_WORDS_IN_BITMAP-1-i), 0);
       }
       tmpmap &= ~onebitmap;
     }
   }
   /* generate events for down keys that were up before */
   for (i=NUM_WORDS_IN_BITMAP-1;i>=0;i--)
   {
     tmpmap = newbitmap[i];
     while ((bitnum = ffs(tmpmap)) != 0)
     {
        onebitmap = 1<<(bitnum-1);
        if ((oldbitmap[i] & onebitmap) == 0) 
        {
	  kp_input_report_key(bitnum + 32*(NUM_WORDS_IN_BITMAP-1-i), 1);
        }
        tmpmap &= ~onebitmap;
     }
   }
   return 0;
}

static void
htcuniversal_kp_report_key(struct key *k, int pressed)
{
	input_report_key(htcuniversal_kp_input_dev, k->code, pressed);
	input_sync(htcuniversal_kp_input_dev);

	if (pressed)
		dprintk("%s pressed\n", k->name);
	else
		dprintk("%s released\n", k->name);
}

static int asic3a_old;
static int asic3b_old;
static int asic3c_old;
static int asic3d_old;

static void
htcuniversal_kp_check_keys(int irq_type, int *old, int new, struct key *k)
{
	int xor;

	xor=new ^ *old;
	if (xor & (1 << k->irq)) 
	{
		htcuniversal_kp_report_key(k, new & (1 << k->irq));
	}
	*old=new;
}

static irqreturn_t htcuniversal_qkbd_irq(int irq, void *data, struct pt_regs *regs)
{
	unsigned long flags;	
	unsigned long kpc_val;

	dprintk("%s: got interrupt for %d...\n", __FUNCTION__, irq);
	input_regs(htcuniversal_kp_input_dev, regs);

        del_timer (&autorepeat_timer);
        spin_lock_irqsave(&keypad_lock,flags);

        /* ack interrupt */
        kpc_val = KPC;

        /* matrix interrupt */
        if (kpc_val & KPC_MI) 
        {
/*
 * The Intel driver turned on KPC_AS here.  It doesn't seem that we
 * would need to do that, because getting an interrupt means that
 * a scan was just done.  For now, I've commented out the setting
 * and clearing of this bit.
           KPC |= KPC_AS;
 */
           while (KPAS & KPAS_SO)
           { 
              /* Wait for the Scan On bit to go off before */
              /* reading the scan registers. */
              NULL;
           };

           kpas = KPAS;
           kpasmkp[0] = KPASMKP0;
           kpasmkp[1] = KPASMKP1;
           kpasmkp[2] = KPASMKP2;
           kpasmkp[3] = KPASMKP3;
          
        }

        /* direct interrupt */
        if (kpc_val & KPC_DI) 
        {
            kpdk = KPDK;
	    /* 
	     * reading the register turns off the "key pressed since last read"
             * bit if it was on, so we turn it off 
	     */
            kpdk &= ~KPDK_DKP;
        }
#ifdef CONFIG_PANIC_LOG
        if ((kpdk & KPDK_DK0)&&(kpasmkp1& 0x00020000))
        {
            if(0 == panic_bug_used)
            {
                panic_bug_used = 1;
                BUG();
            }
        }
#endif
        copy_bitmap (oldkeybitmap, keybitmap);
        scan_to_bitmap (kpas, kpasmkp, keybitmap);

        (void) add_events (oldkeybitmap, keybitmap);

        /* If any keys are down, set a timer to check for key release  */
	/* and one for autorepeat if that's on.                        */
        if (any_keys_down (keybitmap))
        {
#ifdef USE_RELEASE_TIMER
            key_release_timer.expires = (jiffies + key_release_interval);
            add_timer (&key_release_timer);
#endif
            if (do_autorepeat)
            {
              autorepeat_timer.expires = (jiffies + jiffies_to_first_repeat);
              add_timer (&autorepeat_timer);
            }
        }

        spin_unlock_irqrestore(&keypad_lock, flags);
        wake_up_interruptible (&keypad_wait);
//	printk("keypad interrupt occurred\n");
	return IRQ_HANDLED;
}

void kp_handler_switch (int bits, struct key *k) {
	int value;
	value = (bits & (1<<k->irq)) != 0;
	printk("htcuniversal_kp: %s %d\n", k->name, value);
	input_report_switch(htcuniversal_kp_input_dev, k->code, value);
	input_sync(htcuniversal_kp_input_dev);
}

static irqreturn_t htcuniversal_kp_irq(int irq, void *data, struct pt_regs *regs)
{
	struct key *k=data;
	dprintk("%s: got interrupt for %d...\n", __FUNCTION__, irq);
	input_regs(htcuniversal_kp_input_dev, regs);
	if (k->irq_level == IRQT_FALLING)
		k->irq_level=IRQT_RISING;
	else
		k->irq_level=IRQT_FALLING;
	set_irq_type(irq, k->irq_level);
	switch (k->irq_type) {
	case asic3a:
		if(k->handler != NULL)
			(*k->handler)(ipaq_asic3_read_gpio_status_a(&htcuniversal_asic3.dev), k);
		else
			htcuniversal_kp_check_keys(asic3a, &asic3a_old, ~(ipaq_asic3_read_gpio_status_a(&htcuniversal_asic3.dev)), k);
		break;
	case asic3b:
		if(k->handler != NULL)
			(*k->handler)(ipaq_asic3_read_gpio_status_b(&htcuniversal_asic3.dev), k);
		else
			htcuniversal_kp_check_keys(asic3b, &asic3b_old, ~(ipaq_asic3_read_gpio_status_b(&htcuniversal_asic3.dev)), k);
		break;
	case asic3c:
		if(k->handler != NULL)
			(*k->handler)(ipaq_asic3_read_gpio_status_c(&htcuniversal_asic3.dev), k);
		else
			htcuniversal_kp_check_keys(asic3c, &asic3c_old, ~(ipaq_asic3_read_gpio_status_c(&htcuniversal_asic3.dev)), k);
		break;
	case asic3d:
		if(k->handler != NULL)
			(*k->handler)(ipaq_asic3_read_gpio_status_d(&htcuniversal_asic3.dev), k);
		else
			htcuniversal_kp_check_keys(asic3d, &asic3d_old, ~(ipaq_asic3_read_gpio_status_d(&htcuniversal_asic3.dev)), k);
		break;
	}
        return IRQ_HANDLED;
}

static int htcuniversal_kp_init (void)
{
	int i,irq,key, err;
	int cover_rotate, clamshell, night_sensor;
	struct key *k;

	dprintk("%s\n", __FUNCTION__);
	
	htcuniversal_kp_input_dev=input_allocate_device();
        if (!htcuniversal_kp_input_dev) 
        {
 	 printk( KERN_NOTICE "htcuniversal_kbd: unable to allocation keyboard input\n" );
	 return -ENOMEM;
        }

	htcuniversal_kp_input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_REP) | BIT(EV_SW);

	for (i = 0 ; i < num_keys ; i++) {
		k=&keys[i];
		irq=-1;
                set_bit(k->code, htcuniversal_kp_input_dev->keybit);
		switch (k->irq_type) {
		case asic3a:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOA_IRQ_BASE + k->irq;
			break;
		case asic3b:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOB_IRQ_BASE + k->irq;
			break;
		case asic3c:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOC_IRQ_BASE + k->irq;
			break;
		case asic3d:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + k->irq;
			break;
		}
		if (irq != -1) {
			dprintk("requesting irq %d\n", irq);
			request_irq(irq, htcuniversal_kp_irq, SA_INTERRUPT, k->name, k);
			set_irq_type(irq, k->irq_level);
		}
#if 0
		htcuniversal_kp_input_dev.keybit[k->key] |= BIT(k->key);
#endif
	}
	for (i = 0 ; i < num_qkeys ; i++) {
		key=qkeys[i].key & ~SHIFT_BIT;

		if (key) {
			dprintk("enabling key %d\n", qkeys[i].key);
			set_bit(key, htcuniversal_kp_input_dev->keybit);
		}
	}
	set_bit(KEY_RIGHTSHIFT, htcuniversal_kp_input_dev->keybit);

    set_bitmap_to_zero (oldkeybitmap);
    set_bitmap_to_zero (keybitmap);

/* set up gpio */
    pxa_gpio_mode(22 | GPIO_ALT_FN_1_OUT);  /* KP_MKOUT<7> */
    pxa_gpio_mode(36 | GPIO_ALT_FN_3_IN);   /* KP_MKIN<7> */
    pxa_gpio_mode(37 | GPIO_ALT_FN_3_IN);   /* KP_MKIN<3> */
    pxa_gpio_mode(38 | GPIO_ALT_FN_2_IN);   /* KP_MKIN<4> */
    pxa_gpio_mode(40 | GPIO_ALT_FN_1_OUT);  /* KP_MKOUT<6> */
    pxa_gpio_mode(90 | GPIO_ALT_FN_1_IN);   /* KP_MKIN<5> */
    pxa_gpio_mode(91 | GPIO_ALT_FN_1_IN);   /* KP_MKIN<6> */
//    pxa_gpio_mode(93 | GPIO_ALT_FN_1_IN);   /* KP_DKIN<0>, voice_rec */
    pxa_gpio_mode(100 | GPIO_ALT_FN_1_IN);  /* KP_MKIN<0> */
    pxa_gpio_mode(101 | GPIO_ALT_FN_1_IN);  /* KP_MKIN<1> */
    pxa_gpio_mode(102 | GPIO_ALT_FN_1_IN);  /* KP_MKIN<2> */
    pxa_gpio_mode(103 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<0> */
    pxa_gpio_mode(104 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<1> */
    pxa_gpio_mode(105 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<2> */
    pxa_gpio_mode(106 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<3> */
    pxa_gpio_mode(107 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<4> */
    pxa_gpio_mode(108 | GPIO_ALT_FN_2_OUT); /* KP_MKOUT<5> */

    /* set keypad control register */ 
    KPC = (KPC_ASACT |         /* automatic scan on activity */
           KPC_ME |    /* matrix keypad enabled */
//MV           KPC_ME | KPC_DE |   /* matrix and direct keypad enabled */
           (7<<23) | /* 8 columns */
           (7<<26) | /* 8 rows */
           (0<<6) |  /* 1# direct keys */
           KPC_MS_ALL); /* scan all columns */

    pxa_set_cken(CKEN19_KEYPAD,1);

    irq=IRQ_KEYPAD;
    err = request_irq(irq, htcuniversal_qkbd_irq, SA_INTERRUPT, "htcuniversal_qkbd", NULL);
    if (err)
    {
      printk (KERN_CRIT "can't register IRQ%d for keypad, error %d\n",
	      IRQ_KEYPAD, err);
      pxa_set_cken(CKEN19_KEYPAD,0);
      return -ENODEV;
    }
    set_irq_type(irq, IRQT_FALLING);

#ifdef USE_RELEASE_TIMER
    init_timer (&key_release_timer);
    key_release_timer.function = release_timer_went_off;
#endif
    init_timer (&autorepeat_timer);
    autorepeat_timer.function = autorepeat_timer_went_off;

    /* 
     * we want the phone to be able to tell the status of the screen
     * lock switch at power-up time
     */
    kpdk = KPDK;         /* read direct key register */
    /*
     * reading the register turns off the "key pressed since last read" bit
     * if it was on, so we turn it off 
     */
    kpdk &= ~KPDK_DKP;
#if 0
    if (kpdk & KPDK_DK0) /* voice/camera key is pressed */
    {
        turn_on_bit (keybitmap, KEYPAD_CAMERA_VOICE);
    }
#endif

        htcuniversal_kp_input_dev->name = "htcuniversal_kb";
        htcuniversal_kp_input_dev->phys = "keyboard/htcuniversal_kb";

//        htcuniversal_kp_input_dev.id.bustype = BUS_PXA_NSSP;
        htcuniversal_kp_input_dev->id.vendor = 0x0001;
        htcuniversal_kp_input_dev->id.product = 0x0001;
        htcuniversal_kp_input_dev->id.version = 0x0100;

	set_bit(SW_0, htcuniversal_kp_input_dev->swbit);
	set_bit(SW_1, htcuniversal_kp_input_dev->swbit);

        input_register_device(htcuniversal_kp_input_dev);

	KPC |= (KPC_DIE | KPC_MIE);
	KPKDI = 0x40;

/* MV */

	cover_rotate = (ipaq_asic3_read_gpio_status_a(&htcuniversal_asic3.dev) & (1<<GPIOA_COVER_ROTATE_N)) != 0;
	clamshell = (ipaq_asic3_read_gpio_status_b(&htcuniversal_asic3.dev) & (1<<GPIOB_CLAMSHELL_N)) != 0;
	night_sensor = (ipaq_asic3_read_gpio_status_b(&htcuniversal_asic3.dev) & (1<<GPIOB_NIGHT_SENSOR)) != 0;

	printk("htcuniversal_kp: cover_rotate %d\n", cover_rotate);
	printk("htcuniversal_kp: clamshell  %d\n", clamshell);
	printk("htcuniversal_kp: night_sensor  %d\n", night_sensor);

	input_report_switch(htcuniversal_kp_input_dev, SW_0, cover_rotate);
	input_report_switch(htcuniversal_kp_input_dev, SW_1, clamshell);
	input_report_switch(htcuniversal_kp_input_dev, SW_2, night_sensor);

	input_sync(htcuniversal_kp_input_dev);

	return 0;
}

static void htcuniversal_kp_exit (void)
{
	int i,irq;
	struct key *k;
	
	dprintk("%s\n", __FUNCTION__);
	input_unregister_device(htcuniversal_kp_input_dev);
	for (i = 0 ; i < num_keys ; i++) {
		k=&keys[i];
		irq=-1;
		switch (k->irq_type) {
		case asic3a:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOA_IRQ_BASE + k->irq;
			break;
		case asic3b:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOB_IRQ_BASE + k->irq;
			break;
		case asic3c:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOC_IRQ_BASE + k->irq;
			break;
		case asic3d:
			irq=asic3_irq_base(&htcuniversal_asic3.dev) + ASIC3_GPIOD_IRQ_BASE + k->irq;
			break;
		}
		if (irq != -1) {
			dprintk("releasing irq %d\n", irq);
			free_irq(irq, k);
		}
	}
	pxa_set_cken(CKEN19_KEYPAD,0);
	free_irq(IRQ_KEYPAD, NULL);
}

MODULE_AUTHOR("Milan Votava");
MODULE_DESCRIPTION("HTC Universal keypad driver");
MODULE_LICENSE("GPL");

module_init (htcuniversal_kp_init);
module_exit (htcuniversal_kp_exit);

