/*
 * AK4641 codec device driver
 *
 * Copyright (c) 2005 SDG Systems, LLC
 *
 * Based on code:
 *   Copyright (c) 2002 Hewlett-Packard Company
 *   Copyright (c) 2000 Nicolas Pitre <nico@cam.org>
 *   Copyright (c) 2000 Lernout & Hauspie Speech Products, N.V.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/soundcard.h>
#include <linux/i2c.h>
#include <linux/delay.h>

#include <asm/uaccess.h>
#include <asm/mach-types.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/initval.h>
#include <sound/info.h>
#include <sound/ak4641.h>

#define REG_PWR1    0x00
#define REG_PWR2    0x01
#define REG_SEL1    0x02
#define REG_SEL2    0x03
#define REG_MODE1   0x04
#define REG_MODE2   0x05
#define REG_DAC     0x06
#define REG_MIC     0x07
#define REG_TIMER   0x08
#define REG_ALC1    0x09
#define REG_ALC2    0x0a
#define REG_PGA     0x0b
#define REG_LATT    0x0c
#define REG_RATT    0x0d
#define REG_VOL     0x0e
#define REG_STATUS  0x0f
#define REG_EQLO    0x10
#define REG_EQMID   0x11
#define REG_EQHI    0x12
#define REG_BTIF    0x13
#define REG_MAX     REG_BTIF+1

struct ak4641 {
    unsigned char   regs[REG_MAX];
    int             active;
    unsigned short  volume;
    unsigned short  pcm;
    unsigned short  bass;
    unsigned short  treble;
    unsigned short  line;
    unsigned short  mic;
    int             mic_connected;
    int             line_connected;
    int             mod_cnt;
    struct          i2c_client client;
};

struct i2c_client *ak4641_i2c_client;

int ak4641_update(struct i2c_client *clnt, int cmd, void *arg);

#define REC_MASK        (SOUND_MASK_LINE | SOUND_MASK_MIC)
#define DEV_MASK        (REC_MASK | SOUND_MASK_PCM | SOUND_MASK_BASS)

#define REG_PWR1_PMADC  0x01
#define REG_PWR1_PMMIC  0x02
#define REG_PWR1_PMAUX  0x04
#define REG_PWR1_PMMO   0x08
#define REG_PWR1_PMLO   0x10
/* unused               0x20 */
/* unused               0x40 */
#define REG_PWR1_PMVCM  0x80

#define REG_PWR2_PMDAC  0x01
/* unused               0x02 */
/* unused               0x04 */
#define REG_PWR2_PMMO2  0x08
#define REG_PWR2_MCKAC  0x10
/* unused               0x20 */
/* unused               0x40 */
#define REG_PWR2_MCKPD  0x80

#define REG_SEL1_PSMO2  0x01
/* unused               0x02 */
/* unused               0x04 */
/* unused               0x08 */
#define REG_SEL1_MICM   0x10
#define REG_SEL1_DACM   0x20
#define REG_SEL1_PSMO   0x40
#define REG_SEL1_MOGN   0x80

#define REG_SEL2_PSLOR  0x01
#define REG_SEL2_PSLOL  0x02
#define REG_SEL2_AUXSI  0x04
/* unused               0x08 */
#define REG_SEL2_MICL   0x10
#define REG_SEL2_AUXL   0x20
/* unused               0x40 */
#define REG_SEL2_DACL   0x80

#define REG_MODE1_DIF0  0x01
#define REG_MODE1_DIF1  0x02
/* unused               0x04 */
/* unused               0x08 */
/* unused               0x10 */
/* unused               0x20 */
/* unused               0x40 */
/* unused               0x80 */

/* unused               0x01 */
#define REG_MODE2_LOOP  0x02
#define REG_MODE2_HPM   0x04
/* unused               0x08 */
/* unused               0x10 */
#define REG_MODE2_MCK0  0x20
#define REG_MODE2_MCK1  0x40
/* unused               0x80 */

#define REG_DAC_DEM0    0x01
#define REG_DAC_DEM1    0x02
#define REG_DAC_EQ      0x04
/* unused               0x08 */
#define REG_DAC_DATTC   0x10
#define REG_DAC_SMUTE   0x20
#define REG_DAC_TM      0x40
/* unused               0x80 */

#define REG_MIC_MGAIN   0x01
#define REG_MIC_MSEL    0x02
#define REG_MIC_MICAD   0x04
#define REG_MIC_MPWRI   0x08
#define REG_MIC_MPWRE   0x10
#define REG_MIC_AUXAD   0x20
/* unused               0x40 */
/* unused               0x80 */

#define REG_TIMER_LTM0  0x01
#define REG_TIMER_LTM1  0x02
#define REG_TIMER_WTM0  0x04
#define REG_TIMER_WTM1  0x08
#define REG_TIMER_ZTM0  0x10
#define REG_TIMER_ZTM1  0x20
/* unused               0x40 */
/* unused               0x80 */

#define REG_ALC1_LMTH   0x01
#define REG_ALC1_RATT   0x02
#define REG_ALC1_LMAT0  0x04
#define REG_ALC1_LMAT1  0x08
#define REG_ALC1_ZELM   0x10
#define REG_ALC1_ALC1   0x20
/* unused               0x40 */
/* unused               0x80 */

/* REG_EQ controls use 4 bits for each of 5 EQ levels */

/* Bluetooth not implemented */
#define REG_BTIF_PMAD2  0x01
#define REG_BTIF_PMDA2  0x02
#define REG_BTIF_PMBIF  0x04
#define REG_BTIF_ADC2   0x08
#define REG_BTIF_DAC2   0x10
#define REG_BTIF_BTFMT0 0x20
#define REG_BTIF_BTFMT1 0x40
/* unused               0x80 */

static struct ak4641_reg_info {
    unsigned char num;
    unsigned char default_value;
} ak4641_reg_info[] = {
    { REG_PWR1, REG_PWR1_PMVCM | REG_PWR1_PMADC | REG_PWR1_PMLO },
    { REG_PWR2, REG_PWR2_PMMO2 | REG_PWR2_PMDAC },
    { REG_SEL1, REG_SEL1_DACM | REG_SEL1_PSMO2 },
    { REG_SEL2, REG_SEL2_PSLOL | REG_SEL2_PSLOR | REG_SEL2_DACL },
    { REG_MODE1, REG_MODE1_DIF1 },
    { REG_MODE2, 0x0 },
    { REG_DAC, REG_DAC_DEM0 | REG_DAC_DATTC },
    { REG_MIC, REG_MIC_MICAD | REG_MIC_MGAIN },
    { REG_TIMER, 0 },
    { REG_ALC1, 0 },
    { REG_ALC2, 0x36 },
    { REG_PGA, 0x10 },
    { REG_RATT, 0x0 },
    { REG_LATT, 0x0 },
    { REG_VOL, 0x53 },
    { REG_STATUS, 0 },
    { REG_EQLO, 0x88 },
    { REG_EQMID, 0x88 },
    { REG_EQHI, 0x08 },
    { REG_BTIF, REG_BTIF_ADC2 }
};

static u8
ak4641_read_reg( struct i2c_client *clnt, int regaddr )
{
    char buffer[2];
    int r;
    buffer[0] = regaddr;

    r = i2c_master_send(clnt, buffer, 1);
    if (r != 1) {
        printk(KERN_ERR "ak4641: write failed, status %d\n", r);
        return -1;
    }

    r = i2c_master_recv(clnt, buffer, 1);
    if (r != 1) {
        printk(KERN_ERR "ak4641: read failed, status %d\n", r);
        return -1;
    }
    // printk( "ak4641: read addr %x, value %x\n", regaddr, (int)buffer[0] );

    return buffer[0];
}

static void
ak4641_write_reg( struct i2c_client *clnt, int regaddr, int data )
{
    char buffer[2];
    int r;

    buffer[0] = regaddr;
    buffer[1] = data;

#if 0
    printk("ak4641: write %x %x\n", regaddr, data);
#endif
    
    r = i2c_master_send(clnt, buffer, 2);
    if (r != 2) {
        printk(KERN_ERR "ak4641: write failed, status %d\n", r);
    }
}

static void
ak4641_sync( struct i2c_client *clnt )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);
    int i;

#ifdef DEBUG_VERBOSE
    for (i = 0; i < REG_MAX; i++) {
        int v = ak4641_read_reg(clnt, i);
        printk("register %x: old value %x\n", i, v);
    }
#endif

    for (i = 0; i < ARRAY_SIZE(ak4641_reg_info); i++)
        ak4641_write_reg(clnt, ak4641_reg_info[i].num, akm->regs[i]);

#ifdef DEBUG_VERBOSE
    for (i = 0; i < REG_MAX; i++) {
        int v = ak4641_read_reg(clnt, i);
        printk("register %x: new value %x\n", i, v);
    }
#endif
}

#if 0
static void
ak4641_power_up( struct i2c_client *clnt )
{
    ak4641_write_reg(clnt, REG_PWR1, ak4641_reg_info[REG_PWR1].default_value);
    ak4641_write_reg(clnt, REG_PWR2, ak4641_reg_info[REG_PWR2].default_value);
}
#endif

static void
ak4641_power_down( struct i2c_client *clnt )
{
    ak4641_write_reg(clnt, REG_PWR1, 0);
    /* MCKPD - disable MCLK input buffer control */
    ak4641_write_reg(clnt, REG_PWR2, REG_PWR2_MCKPD);
}

static void
ak4641_cmd_init( struct i2c_client *clnt )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);

    akm->active = 1;
    // ak4641_power_up(clnt);
    /* resend all parameters */
    ak4641_sync(clnt);
}

static int
ak4641_configure( struct i2c_client *clnt, struct ak4641_cfg *conf )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);
    int ret = 0;

    // printk("ak4641_configure: mic_connected=%d\n", conf->mic_connected);

    akm->mic_connected = conf->mic_connected;
    akm->line_connected = 0;

    return ret;
}

int
ak4641_update( struct i2c_client *clnt, int cmd, void *arg )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);
    unsigned char newregnum = 255;
    unsigned char altregnum = 255;
    unsigned char tmp;
    unsigned char val = *(int *)arg;

    switch (cmd) {
        case I2C_SET_PCM:
            newregnum = REG_VOL;
            if (val > 15) {
                printk("pcmvol=%d must be in range [0,15]; using 15", val);
                val = 15;
            }
            tmp = akm->regs[REG_VOL] & 0xf0;
            tmp |= val;
            akm->regs[REG_VOL] = tmp;
            break;

        case I2C_SET_BASS:
            newregnum = REG_DAC;
            if (val > 3) {
                printk("bass=%d must be in range [0,3]; using 3", val);
                val = 3;
            }
            tmp = akm->regs[REG_DAC] & ~3;
            tmp |= val;
            akm->regs[REG_DAC] = tmp;
            break;

        case I2C_SET_MUTE:
            newregnum = REG_DAC;
            if (val > 0) /* mute-on is non-zero */
                tmp = akm->regs[REG_DAC] | REG_DAC_SMUTE;
            else
                tmp = akm->regs[REG_DAC] & ~REG_DAC_SMUTE;
            akm->regs[REG_DAC] = tmp;
            break;

        case I2C_SET_TREBLE: /* set treble.  val = 50 to 100 => 0 to 3 */
            /* not supported */
            return -EINVAL;

        case I2C_SET_LINE_GAIN:
            /* 0..255 valid, no check for val needed */
            newregnum = REG_LATT;
            akm->regs[REG_LATT] = val;
            altregnum = REG_RATT;
            akm->regs[REG_RATT] = val;
            break;

        case I2C_SET_MIC_GAIN:
            newregnum = REG_PGA;
            if (val > 127) {
                printk("mic_gain=%d must be in range [0,127]; using 127", val);
                val = 127;
            }
            akm->regs[REG_PGA] = val;
            break;

        case I2C_SET_AGC:
            break;

        case I2C_SET_RECSRC_LINE:
            newregnum = REG_MIC;
	    altregnum = REG_PWR1;
            akm->regs[REG_MIC] |= REG_MIC_MSEL | REG_MIC_MPWRE;
            akm->regs[REG_MIC] &= ~REG_MIC_MPWRI;
	    akm->regs[REG_PWR1] &= ~REG_PWR1_PMMIC;
            break;

        case I2C_SET_RECSRC_MIC:
            altregnum = REG_MIC;
	    newregnum = REG_PWR1;
	    if (val != 0) {
		akm->regs[REG_MIC] &= ~(REG_MIC_MPWRE | REG_MIC_MSEL);
		akm->regs[REG_MIC] |= REG_MIC_MPWRI;
		akm->regs[REG_PWR1] |= REG_PWR1_PMMIC;
	    }
	    else {
		akm->regs[REG_MIC] &= ~(REG_MIC_MPWRI);
		akm->regs[REG_PWR1] &= ~REG_PWR1_PMMIC;
	    }
            break;

        case I2C_SET_HEADPHONE_SWITCH:
            newregnum = REG_PWR1;
            if (val) {
                akm->regs[REG_PWR1] |= REG_PWR1_PMLO;
            }
            else {
                akm->regs[REG_PWR1] &= ~REG_PWR1_PMLO;
            }
            break;

        case I2C_SET_HEADPHONE_OUTPUT:
            newregnum = REG_SEL2;
            if (val) {
                akm->regs[REG_SEL2] |= REG_SEL2_PSLOL | REG_SEL2_PSLOR;
            }
            else {
                akm->regs[REG_SEL2] &= ~(REG_SEL2_PSLOL | REG_SEL2_PSLOR);
            }
            break;

        default:
            return -EINVAL;
    }       

    if (akm->active) {
	// printk( "ak4641_update writing %d and %d\n", newregnum, altregnum );
        ak4641_write_reg(clnt, newregnum, akm->regs[newregnum]);
        if (altregnum != 255)
            ak4641_write_reg(clnt, altregnum, akm->regs[altregnum]);
    }
    return 0;
}

static int
ak4641_mixer_ioctl( struct i2c_client *clnt, int cmd, void *arg )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);
    struct i2c_gain gain;
    int val, nr = _IOC_NR(cmd), ret = 0;

    // printk( "ak4641_mixer_ioctl: %d\n", cmd );
    if (cmd == SOUND_MIXER_INFO) {
        struct mixer_info mi;

        strncpy(mi.id, "AK4641", sizeof(mi.id));
        strncpy(mi.name, "Asahi Kasei AK4641", sizeof(mi.name));
        mi.modify_counter = akm->mod_cnt;
        return copy_to_user(arg, &mi, sizeof(mi));
    }

    else if (cmd == SOUND_MIXER_WRITE_RECSRC) {
        ret = get_user(val, (int *)arg);

        if (ret)
            goto out;

        if (val & SOUND_MASK_LINE) {
            akm->line_connected = 1;
            akm->mic_connected = 0;
            ak4641_update(clnt, I2C_SET_RECSRC_LINE, arg);
            ak4641_update(clnt, I2C_SET_HEADPHONE_SWITCH,
                &akm->line_connected);
            ak4641_update(clnt, I2C_SET_HEADPHONE_OUTPUT,
                &akm->line_connected);
        }
        if (val & SOUND_MASK_MIC) {
            akm->mic_connected = 1;
            akm->line_connected = 0;
            ak4641_update(clnt, I2C_SET_RECSRC_MIC, arg);
            ak4641_update(clnt, I2C_SET_HEADPHONE_SWITCH,
                &akm->line_connected);
            ak4641_update(clnt, I2C_SET_HEADPHONE_OUTPUT,
                &akm->line_connected);
        }
        return put_user(val, (int *)arg);
    }

    else if (cmd == SOUND_MIXER_READ_RECSRC) {
        val = 0;
        if (akm->line_connected) val |= SOUND_MASK_LINE;
        if (akm->mic_connected) val |= SOUND_MASK_MIC;
        return put_user(val, (int *) arg);
    }

    if (_IOC_DIR(cmd) & _IOC_WRITE) {
        ret = get_user(val, (int *)arg);
        if (ret)
            goto out;

        gain.left    = val & 255;       /* 127? */
        gain.right   = val >> 8;

        val &= 255;     /* XXX? max largest value */

        switch (nr) {
            case SOUND_MIXER_VOLUME:
            case SOUND_MIXER_PCM:
                akm->pcm = val;         /* XXX set to val-adjusted value? */
                akm->volume = val;      /* ditto? */
                akm->mod_cnt++;
                /* input is 0-100, ak4641 uses 0-15 */
                val = val * 15  / 100;
                ak4641_update(clnt, I2C_SET_PCM, &val);
                break;

            case SOUND_MIXER_BASS:
                akm->bass = val;
                akm->mod_cnt++;
                val = val * 3 / 100;
                ak4641_update(clnt, I2C_SET_BASS, &gain);
                break;

            case SOUND_MIXER_LINE:
                if (!akm->line_connected)
                    return -EINVAL;
                akm->line = val;
                akm->mod_cnt++;
                ak4641_update(clnt, I2C_SET_LINE_GAIN, &val);
                break;

            case SOUND_MIXER_MIC:
                if (!akm->mic_connected)
                    return -EINVAL;
                akm->mic = val;
                akm->mod_cnt++;
                /* input value 0-100, ak4641 uses 0-127 */
                val = val * 127 / 100;
                ak4641_update(clnt, I2C_SET_MIC_GAIN, &gain);
                break;

            case SOUND_MIXER_RECSRC:
                break;

            case SOUND_MIXER_AGC:
                ak4641_update(clnt, I2C_SET_AGC, &val);
                break;

            default:
                ret = -EINVAL;
        }
    }

    if (ret == 0 && _IOC_DIR(cmd) & _IOC_READ) {
        int nr = _IOC_NR(cmd);
        ret = 0;

        switch (nr) {
            case SOUND_MIXER_VOLUME:
                val = akm->volume;
                break;
            case SOUND_MIXER_PCM:
                val = akm->pcm;
                break;
            case SOUND_MIXER_BASS:
                val = akm->bass;
                break;
            case SOUND_MIXER_TREBLE:
                val = akm->treble;
                break;
            case SOUND_MIXER_LINE:
                val = akm->line;
                break;
            case SOUND_MIXER_MIC:
                val = akm->mic;
                break;
            case SOUND_MIXER_RECSRC:
                val = REC_MASK;
                goto devmask;
            case SOUND_MIXER_RECMASK:
                val = REC_MASK;
                goto devmask;
            case SOUND_MIXER_DEVMASK:
                val = DEV_MASK;
devmask:
                if (!akm->mic_connected) val &= ~SOUND_MASK_MIC;
                if (!akm->line_connected) val &= ~SOUND_MASK_LINE;
                break;

            case SOUND_MIXER_CAPS:
                val = 0;
                break;
            case SOUND_MIXER_STEREODEVS:
                val = 0;
                break;
            default:
                val = 0;
                ret = -EINVAL;
                break;
        }

        if (ret == 0)
            ret = put_user(val, (int *)arg);
    }
out:
    return ret;
}

/* 
 * Mixer
 */

#define AK4641_REG(xname, reg, shift, mask, invert) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
  .name = xname, \
  .access = SNDRV_CTL_ELEM_ACCESS_READWRITE, \
  .info = snd_ak4641_info_reg, \
  .get = snd_ak4641_get_reg, \
  .put = snd_ak4641_put_reg, \
  .private_value = 1 | (reg << 5) | (shift << 11) | (mask << 14) | (invert << 20) \
}

#define AK4641_REG_DUAL(xname, reg, shift, mask, invert) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, \
  .name = xname, \
  .access = SNDRV_CTL_ELEM_ACCESS_READWRITE, \
  .info = snd_ak4641_info_reg, \
  .get = snd_ak4641_get_reg, \
  .put = snd_ak4641_put_reg, \
  .private_value = 2 | (reg << 5) | (shift << 11) | (mask << 14) | (invert << 20) \
}

int
snd_ak4641_info_reg( snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t * uinfo )
{
    int count = (kcontrol->private_value >> 0) & 3;
    int mask = (kcontrol->private_value >> 14) & 63;

    // printk("snd_ak4641_info_reg: kcontrol=%p:%s private_value=%p uinfo=%p\n", 
        // kcontrol, kcontrol->id.name, (void*)kcontrol->private_value, uinfo);

    uinfo->type = mask == 1 ? SNDRV_CTL_ELEM_TYPE_BOOLEAN : SNDRV_CTL_ELEM_TYPE_INTEGER;
    uinfo->count = count;
    uinfo->value.integer.min = 0;
    uinfo->value.integer.max = mask;

    return 0;
}

int
snd_ak4641_get_reg( snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol )
{
    // snd_card_hx4700_ak4641_t *chip = snd_kcontrol_chip(kcontrol);
    struct i2c_client *clnt = NULL;
    struct ak4641 *akm = NULL;
    int count = (kcontrol->private_value >> 0) & 3;
    int reg = (kcontrol->private_value >> 5) & 63;
    int shift = (kcontrol->private_value >> 11) & 7;
    int mask = (kcontrol->private_value >> 14) & 63;
    int invert = (kcontrol->private_value >> 20) & 1;
    unsigned char regval;
    int i;

    // printk("snd_ak4641_get_reg: kcontrol=%p:%s reg=%d\n",
        // kcontrol, kcontrol->id.name, reg);

    clnt = ak4641_i2c_client;
    if (clnt)
        akm = i2c_get_clientdata(clnt);
    else
        return -1;

    if (akm == NULL || ucontrol == NULL)
	return -1;

    for (i = 0; i < count; i++) {
        regval= akm->regs[reg+i] = ak4641_read_reg(clnt, reg+i);
        
        ucontrol->value.integer.value[i] = (regval >> shift) & mask;
        if (invert)
            ucontrol->value.integer.value[i] = mask - ucontrol->value.integer.value[i];
    }
    return 0;
}

int
snd_ak4641_put_reg( snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol )
{
    // snd_card_hx4700_ak4641_t *chip = snd_kcontrol_chip(kcontrol);
    struct i2c_client *clnt = NULL;
    struct ak4641 *akm = NULL;
    int count = (kcontrol->private_value >> 0) & 3;
    int reg = (kcontrol->private_value >> 5) & 63;
    int shift = (kcontrol->private_value >> 11) & 7;
    int mask = (kcontrol->private_value >> 14) & 63;
    int invert = (kcontrol->private_value >> 20) & 1;
    int i;
    unsigned char val;
    unsigned char regval;

    clnt = ak4641_i2c_client;
    if (clnt)
        akm = i2c_get_clientdata(clnt);

    // printk("snd_ak4641_put_reg: control=%s reg=%d\n",
           // kcontrol->id.name, reg);

    for (i = 0; i < count; i++) {
        regval = akm->regs[reg+i];
        val = (ucontrol->value.integer.value[i] & mask);
        if (invert)
            val = mask - val;
        regval &= ~(mask << shift);
        regval |= (val << shift);

        akm->regs[reg+i] = regval;

        ak4641_write_reg(clnt, reg+i, regval);
    }
    return 0;
}

#define AK4641_CONTROLS (sizeof(snd_ak4641_controls)/sizeof(snd_kcontrol_new_t))

static snd_kcontrol_new_t snd_ak4641_controls[] =
{
    AK4641_REG("ADC Power Switch",      REG_PWR1, 0, 1, 0),
    AK4641_REG("Mic Power Switch",      REG_PWR1, 1, 1, 0),
    AK4641_REG("Aux In Power Switch",       REG_PWR1, 2, 1, 0),
    AK4641_REG("Mono Out Power Switch",     REG_PWR1, 3, 1, 0),
    AK4641_REG("Line Out Power Switch",     REG_PWR1, 4, 1, 0),
    AK4641_REG("VCOM Power Switch",     REG_PWR1, 7, 1, 0),

    AK4641_REG("DAC Power Switch",         REG_PWR2, 0, 1, 0),
    AK4641_REG("Mono Out2 Power Switch",   REG_PWR2, 3, 1, 0),

    AK4641_REG("Mono2 Out Switch", REG_SEL1, 0, 1, 0),
    AK4641_REG("Mono Out Switch", REG_SEL1, 6, 1, 0),
    AK4641_REG("Mono Gain Switch", REG_SEL1, 7, 1, 0),

    AK4641_REG("Right Channel Power Switch", REG_SEL2, 0, 1, 0),
    AK4641_REG("Left Channel Power Switch",  REG_SEL2, 1, 1, 0),
    AK4641_REG("Mic to Stereo Mixer Switch",  REG_SEL2, 4, 1, 0),
    AK4641_REG("Aux to Stereo Mixer Switch",  REG_SEL2, 5, 1, 0),
    AK4641_REG("DAC to Stereo Mixer Switch",  REG_SEL2, 7, 1, 0),

    AK4641_REG("Loopback Switch",  REG_MODE2, 1, 1, 0),

    AK4641_REG("Tone - De-emphases response",   REG_DAC, 0, 3, 0),
    AK4641_REG("Equalizer Switch",   REG_DAC, 2, 1, 0),
    AK4641_REG("Attenuator (L/R) Switch",   REG_DAC, 4, 1, 0),
    AK4641_REG("Soft Mute Switch",   REG_DAC, 5, 1, 0),

    AK4641_REG("Mic 20dB Gain Switch",  REG_MIC, 0, 1, 0),
    AK4641_REG("Int/Ext Mic Select (0=Internal)",  REG_MIC, 1, 1, 0),
    AK4641_REG("Mic Playback Switch",   REG_MIC, 2, 1, 0),
    AK4641_REG("Internal Mic Power Switch", REG_MIC, 3, 1, 0),
    AK4641_REG("External Mic Power Switch", REG_MIC, 4, 1, 0),

    AK4641_REG("Input Volume",   REG_PGA, 0, 127, 0),

    AK4641_REG("PCM Playback Volume (L)",  REG_LATT, 0, 255, 0),
    AK4641_REG("PCM Playback Volume (R)",  REG_RATT, 0, 255, 0),

    AK4641_REG("AUX In Volume",  REG_VOL, 0, 15, 0),
    AK4641_REG("Mic In Attenuator",  REG_VOL, 4, 7, 0),
    AK4641_REG("Mic In -4dB",  REG_VOL, 7, 1, 0),
};

snd_kcontrol_t *snd_ak4641_kcontrol[AK4641_CONTROLS];

int
snd_card_ak4641_mixer_new( snd_card_t *card )
{
    int idx, err;

    // printk("%s:\n", __FUNCTION__);
        
    snd_assert(card != NULL, return -EINVAL);

    // printk("ak4641 adding mixer controls.\n");
    for (idx = 0; idx < AK4641_CONTROLS; idx++) {
	snd_ak4641_kcontrol[idx] = snd_ctl_new1( &snd_ak4641_controls[idx], ak4641_i2c_client );
	if (snd_ak4641_kcontrol[idx] != NULL) {
	    if ((err = snd_ctl_add(card,snd_ak4641_kcontrol[idx])) < 0) {
		return err;
	    }
	}
	else {
	    return -ENOMEM;
	}
    }

#if 0
    /* set default values */
    {
        snd_kcontrol_t * tmp_k = kmalloc(sizeof(snd_kcontrol_t), GFP_KERNEL);
        snd_ctl_elem_value_t * tmp_v = kmalloc(sizeof(snd_ctl_elem_value_t), GFP_KERNEL);

        tmp_k->private_value = CMD_VOLUME;
        tmp_v->value.integer.value[0] = 10;
        snd_ak4641_put_reg(tmp_k, tmp_v);

        kfree(tmp_k);
        kfree(tmp_v);
    }
#endif
    strcpy(card->mixername, "AK4641 Mixer");

    return 0;
}

void
snd_card_ak4641_mixer_del( snd_card_t *card )
{
#if 0
    int idx;

    for (idx=0; idx<AK4641_CONTROLS; idx++) {
	snd_ctl_remove_id( card, snd_ak4641_kcontrol[idx]->id );
    }
#endif
    // printk("ak4641 mixer_del\n");
}

static struct i2c_driver ak4641_driver;	/* forward declaration to one below */

static struct i2c_client client_template = {
    name: "(unset)",
    driver: &ak4641_driver
};

static int
ak4641_attach( struct i2c_adapter *adap, int addr, int zero_or_minus_one )     
{
    struct ak4641 *akm;
    struct i2c_client *clnt;
    int i;

    akm = kmalloc(sizeof(*akm), GFP_KERNEL);
    if (!akm)
        return -ENOMEM;

    memset(akm, 0, sizeof(*akm));

    /* set default values of registers */
    for (i = 0; i < ARRAY_SIZE(ak4641_reg_info); i++) {
        int addr = ak4641_reg_info[i].num;
        akm->regs[addr] = ak4641_reg_info[i].default_value;
    }
    akm->mic_connected = 1;


    clnt = &akm->client;
    memcpy( clnt, &client_template, sizeof *clnt );
    clnt->adapter = adap;
    clnt->addr = addr;
    strcpy( clnt->name, "ak4641" );

    // printk( KERN_NOTICE "attaching client %p\n", clnt );
    ak4641_i2c_client = clnt;   /* XXX part of hack */

    i2c_attach_client( clnt );
    i2c_set_clientdata( clnt, akm );

    return 0;
}

static int
ak4641_detach_client( struct i2c_client *clnt )
{
    i2c_detach_client(clnt);

    kfree(clnt);

    return 0;
}

/* Addresses to scan */
static unsigned short force0[] = {ANY_I2C_BUS, 0x12};
static unsigned short force1[] = {I2C_CLIENT_END, I2C_CLIENT_END};
static unsigned short normal_i2c[] = {0x12,I2C_CLIENT_END};
static unsigned short probe[]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[]       = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short *forces[]        = { force0, force1, 0 };

static struct i2c_client_address_data addr_data = {
    .normal_i2c = normal_i2c,
    .probe = probe,
    .ignore = ignore,
    .forces = forces,
};

static int
ak4641_attach_adapter( struct i2c_adapter *adap )
{
    // printk( KERN_NOTICE "ak4641_attach_adapter\n" );
    return i2c_probe(adap, &addr_data, ak4641_attach);
}

static int
ak4641_open( struct i2c_client *clnt )
{
    ak4641_cmd_init( clnt );
    return 0;
}

static void
ak4641_close( struct i2c_client *clnt )
{
    struct ak4641 *akm = container_of(clnt, struct ak4641, client);

    ak4641_power_down( clnt );
    akm->active = 0;
}

int
ak4641_command( struct i2c_client *clnt, unsigned int cmd, void *arg )
{
    int ret = -EINVAL;

    if (_IOC_TYPE(cmd) == 'M')
        ret = ak4641_mixer_ioctl(clnt, cmd, arg);
    else if (cmd == I2C_AK4641_CONFIGURE)
        ret = ak4641_configure(clnt, arg);
    else if (cmd == I2C_AK4641_OPEN)
        ret = ak4641_open(clnt);
    else if (cmd == I2C_AK4641_CLOSE)
        (ret = 0), ak4641_close(clnt);

    return ret;
}
EXPORT_SYMBOL(ak4641_command);

static struct i2c_driver ak4641_driver = {
    .driver = {
            .name   = AK4641_NAME,
    },
    .id             = I2C_DRIVERID_AK4641,
    .class          = I2C_CLASS_SOUND,
    .attach_adapter = ak4641_attach_adapter,
    .detach_client  = ak4641_detach_client,
    .command        = ak4641_command
};

/*
 * HACK: i2c_get_client is not defined! Use this temporarily
 */
struct i2c_client *
ak4641_i2c_get_client( void )
{
    return ak4641_i2c_client;
}
EXPORT_SYMBOL(ak4641_i2c_get_client);


int
snd_ak4641_activate( void )
{
    int ret;

    if (machine_is_h4700()) {
	// printk( KERN_NOTICE "i2c_add_driver\n" );
        ret = i2c_add_driver( &ak4641_driver );
	if (ret != 0) return ret;
	mdelay(10);
	// printk( KERN_NOTICE "ak4641_open %p\n", ak4641_i2c_client );
        ret = ak4641_open( ak4641_i2c_client );
	return ret;
    }
    else
        return -ENODEV;
}

void
snd_ak4641_deactivate( void )
{
    if (machine_is_h4700()) {
	if (ak4641_i2c_client != NULL)
	    ak4641_close( ak4641_i2c_client );
        i2c_del_driver( &ak4641_driver );
    }
}

int
snd_ak4641_mic( int on )
{
    return ak4641_update( ak4641_i2c_client, I2C_SET_RECSRC_MIC, &on );
}

static snd_card_t *card;

int
snd_ak4641_mixer_create( snd_card_t *_card )
{
    card = _card;
    return snd_card_ak4641_mixer_new( card ); // ??? &ak4641_i2c_client );   
}

void
snd_ak4641_mixer_free( void )
{
    snd_card_ak4641_mixer_del( card );
}

void
snd_ak4641_no_plop( int on )
{
    // printk( "snd_ak4641_no_plop: %d\n", on );
    /* MEMO: add proper state */
    ak4641_update( ak4641_i2c_client, I2C_SET_MUTE, &on );
}

static int __init ak4641_init(void) { return 0; }
static void __exit ak4641_exit(void) {}

module_init(ak4641_init);
module_exit(ak4641_exit);

MODULE_AUTHOR("Todd Blumer <todd@sdgsystems.com>");
MODULE_DESCRIPTION("AK4641 CODEC driver");
MODULE_LICENSE( "GPL" );

EXPORT_SYMBOL(snd_ak4641_activate);
EXPORT_SYMBOL(snd_ak4641_deactivate);
EXPORT_SYMBOL(snd_ak4641_no_plop);
EXPORT_SYMBOL(snd_ak4641_mixer_create);
EXPORT_SYMBOL(snd_ak4641_mixer_free);
EXPORT_SYMBOL(snd_ak4641_mic);

